<?php

/**
 * Zend Framework (http://framework.zend.com/)
 *
 * @link      http://github.com/zendframework/Quantity for the canonical source repository
 * @copyright Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license   http://framework.zend.com/license/new-bsd New BSD License
 */
namespace Quantity\Controller;

use Zend\Mvc\Controller\AbstractRestfulController;
use Zend\View\Model\JsonModel;
use Zend\Http\Response;

class QuantityController extends AbstractRestfulController {

    public function get($productNumber) {
        // Check if authorised
        $isAuthorised = $this->AuthorisationController()->isAuthorisationValid($this->getRequest());
        if(!$isAuthorised) {
            return $this->emptyResponse($this->getResponse(), 403);
        }

        // Create mappers
        /** @var $quantityMapper \Quantity\Model\QuantityMapper */
        $quantityMapper = $this->getMapper('QuantityMapper');

        // Get quantity for product number
        $quantity = $quantityMapper->getQuantityByProductNumber($productNumber);

        // Check quantity is valid (null if product doesn't exist)
        if($quantity === null) {
            return $this->emptyResponse($this->getResponse(), 404);
        }

        return new JsonModel(array(
            'total_quantity' => $quantity['total_quantity'],
            'available_quantity' => $quantity['available_quantity'],
        ));
    }

    public function patch($productNumber, $data) {
        // Check if authorised
        $isAuthorised = $this->AuthorisationController()->isAuthorisationValid($this->getRequest());
        if(!$isAuthorised) {
            return $this->emptyResponse($this->getResponse(), 403);
        }

        // Create mappers
        /** @var $quantityMapper \Quantity\Model\QuantityMapper */
        $quantityMapper = $this->getMapper('QuantityMapper');

        // Get quantity from data
        $quantityToIncreaseBy = (isset($data['quantity']) ? $data['quantity'] : -1);

        // Error if quantity is < 1
        if ($quantityToIncreaseBy < 1) {
            return $this->emptyResponse($this->getResponse(), 400);
        }

        // Save quantity
        $hasIncreased = $quantityMapper->increaseQuantityByProductNumber($productNumber, $quantityToIncreaseBy);

        // If not increased, the product number doesn't exist so return a 404
        if (!$hasIncreased) {
            return $this->emptyResponse($this->getResponse(), 404);
        }

        // Otherwise it was increased succesfully so return the new quantities
        return $this->get($productNumber);
    }

    public function create($data) {
        // Check if authorised
        $isAuthorised = $this->AuthorisationController()->isAuthorisationValid($this->getRequest());
        if(!$isAuthorised) {
            return $this->emptyResponse($this->getResponse(), 403);
        }

        // Create mappers
        /** @var $quantityMapper \Quantity\Model\QuantityMapper */
        $quantityMapper = $this->getMapper('QuantityMapper');

        // Get quantity from data
        $quantityOfProduct = (isset($data['quantity']) ? $data['quantity'] : -1);
        $productNumber = (isset($data['product_number']) ? $data['product_number'] : null);

        // Ensure product doesn't already exist
        $currentQuantity = $quantityMapper->getQuantityByProductNumber($productNumber);
        if($currentQuantity !== null) {
            return $this->emptyResponse($this->getResponse(), 409);
        }

        // Error if quantity is < 0
        if ($quantityOfProduct < 0) {
            return $this->emptyResponse($this->getResponse(), 400);
        }

        // Save quantity
        $isCreated = $quantityMapper->createProductNumberAndSetQuantity($productNumber, $quantityOfProduct);

        // If not increased, something went wrong somewhere so just throw an internal server error
        if (!$isCreated) {
            return $this->emptyResponse($this->getResponse(), 500);
        }

        // Otherwise it was created succesfully so return the new quantities
        return $this->get($productNumber);
    }

    private function emptyResponse(Response $response, $statusCode) {
        $response->setContent('');
        return $response->setStatusCode($statusCode);
    }

    private function getMapper($mapperName) {
        return $this->getServiceLocator()->get($mapperName);
    }
}
