<?php

namespace Quantity\Model;

use Zend\Db\Sql\Sql;
use Zend\Stdlib\Hydrator\ClassMethods;
use Zend\Db\ResultSet\HydratingResultSet;

class QuantityMapper {

    protected $sql;

    public function __construct(Sql $sql) {
        $this->sql = $sql;
    }

    /**
     * Gets the stock level values for all products.
     *
     * @return Zend\Db\ResultSet\HydratingResultSet iterable of Quantity\Model\StockLevelEntity classes
     */
    public function getAllQuantities() {
        $select = $this->sql->select('product_quantity');
        $select->order(array('product_number ASC'));

        $statement = $this->sql->prepareStatementForSqlObject($select);
        $results = $statement->execute();

        $entityPrototype = new StockLevelEntity();
        $hydrator = new ClassMethods();
        $resultset = new HydratingResultSet($hydrator, $entityPrototype);
        $resultset->initialize($results);
        $resultset->buffer();

        return $resultset;
    }

    /**
     * Gets the total and available quantity for a particular product.
     *
     * @param string $productNumber the product number of the item to retrieve quantity for
     *
     * @return array|NULL null if the given product number doesn't exist, otherwise an array with keys total_quantity, reserved_quantity, and available_quantity
     */
    public function getQuantityByProductNumber($productNumber) {
        $select = $this->sql->select('product_quantity');
        $select->where->equalTo('product_number', $productNumber);
        $select->forUpdate = true;

        $statement = $this->sql->prepareStatementForSqlObject($select);
        $results = $statement->execute()->current();

        if (!$results) {
            return null;
        }

        return array(
            'total_quantity' => $results['total_quantity'],
            'reserved_quantity' => $results['reserved_quantity'],
            'available_quantity' => ($results['total_quantity'] - $results['reserved_quantity'])
        );
    }

    /**
     * Increases the total quantity of a product by a given amount.
     *
     * @param string $productNumber the product number of the item whose quantity should be increased
     * @param int $quantityToIncreaseBy the amount to increase the product's stock levels by
     *
     * @return boolean true if quantity has been increased, false if not or if the product number doesn't exist
     */
    public function increaseQuantityByProductNumber($productNumber, $quantityToIncreaseBy) {
        // Start transaction
        $connection = $this->sql->getAdapter()->getDriver()->getConnection();
        $connection->beginTransaction();

        try {

            // Get current quantity
            $currentQuantity = $this->getQuantityByProductNumber($productNumber);

            if (!$currentQuantity) {
                $connection->rollback();
                return false;
            }

            $update = $this->sql->update('product_quantity');
            $update->set(array(
                'total_quantity' => ($currentQuantity['total_quantity'] + $quantityToIncreaseBy)
            ));
            $update->where->equalTo('product_number', $productNumber);

            $statement = $this->sql->prepareStatementForSqlObject($update);
            $statement->execute();

            $connection->commit();

            return true;

        } catch (\Exception $e) {
            $connection->rollback();
            return false;
        }
    }

    /**
     * Creates a new product and sets its quantity.
     *
     * @param string $productNumber the product number of the item to be added
     * @param int $quantityOfProduct the initial quantity of stock to set the product's stock levels to
     *
     * @return boolean true if the product has been added and the stock quantity has been set, or false if the product already exists
     */
    public function createProductNumberAndSetQuantity($productNumber, $quantityOfProduct) {
        // Start transaction
        $connection = $this->sql->getAdapter()->getDriver()->getConnection();
        $connection->beginTransaction();

        try {

            // Get current quantity
            $currentQuantity = $this->getQuantityByProductNumber($productNumber);

            if ($currentQuantity !== null) {
                $connection->rollback();
                return false;
            }

            $insert = $this->sql->insert('product_quantity');
            $insert->values(array(
                'product_number' => $productNumber,
                'total_quantity' => $quantityOfProduct,
                'reserved_quantity' => 0
            ));

            $statement = $this->sql->prepareStatementForSqlObject($insert);
            $statement->execute();

            $connection->commit();

            return true;

        } catch (\Exception $e) {
            $connection->rollback();
            return false;
        }
    }
}

?>