<?php

namespace Reservation\Model;

use Zend\Db\Sql\Sql;
use Quantity\Model\QuantityMapper;

class ReservationMapper {

    protected $sql;
    private $TOKEN_EXPIRY_IN_MINUTES = 15;

    public function __construct(Sql $sql) {
        $this->sql = $sql;
    }

    /**
     * Gets the expiry date and time of a reservation token.
     *
     * @param string $reservationToken reservation token to check the expiry of
     *
     * @return NULL|string null if the reservation token doesn't exist, otherwise the date and time that the token expires in the format YYYY-MM-DD HH:MM:SS
     */
    public function getReservationTokenExpiry($reservationToken) {
        $select = $this->sql->select('product_reservation_token');
        $select->where->equalTo('token', $reservationToken);

        $statement = $this->sql->prepareStatementForSqlObject($select);
        $results = $statement->execute()->current();

        if (!$results) {
            return null;
        }

        if ($results['expiration_processed'] == 1 && $results['token_expiry'] >= date('Y-m-d H:i:s')) {
            return date('Y-m-d H:i:s', strtotime('- 1 minute'));
        }

        return $results['token_expiry'];
    }

    /**
     * Gets the currently reserved quantity for a given product and reservation token.
     *
     * @param string $productNumber the product number to get the reserved quantity for
     * @param string $reservationToken the reservation token to get the reserved quantity for
     *
     * @return int the current quantity of the product that has been reserved for the reservation token
     */
    public function getReservedQuantityByProductNumberAndReservationToken($productNumber, $reservationToken) {
        $select = $this->sql->select('product_reservation');
        $select->join('product_reservation_token', 'product_reservation_token.id = product_reservation.product_reservation_token_id', array());
        $select->where->equalTo('product_reservation.product_number', $productNumber);
        $select->where->equalTo('product_reservation_token.token', $reservationToken);
        $select->forUpdate = true;

        $statement = $this->sql->prepareStatementForSqlObject($select);
        $results = $statement->execute()->current();

        if (!$results) {
            return 0;
        }

        return $results['reserved_quantity'];
    }

    /**
     * Sets the reserved quantity of a product for a reservation token.
     *
     * @param string $productNumber the product number to set the reserved quantity of
     * @param int $requestedQuantity the quantity of the product that has been requested to be reserved
     * @param string $reservationToken the reservation token to make the reservation under
     * @param QuantityMapper $quantityMapper the QuantityMapper class, used to get current stock levels
     *
     * @return boolean true if the reservation has been made successfully, or false if there isn't sufficient stock to fulfil the request
     */
    public function setReservationQuantity($productNumber, $requestedQuantity, $reservationToken, QuantityMapper $quantityMapper) {
        // Start transaction
        $connection = $this->sql->getAdapter()->getDriver()->getConnection();
        $connection->beginTransaction();

        try {
            // Get current reserved quantity
            $currentReservedQuantity = $this->getReservedQuantityByProductNumberAndReservationToken($productNumber, $reservationToken);

            // Get current quantity
            $currentQuantity = $quantityMapper->getQuantityByProductNumber($productNumber);

            // Get reservation token ID
            $reservationTokenId = $this->getReservationTokenIdByReservationToken($reservationToken);

            // If the quantity requested is the current reserved quantity then there's nothing to do
            if ($requestedQuantity == $currentReservedQuantity) {
                return true;
            }

            // Otherwise calculate how much to change the reservation quantity by
            $reservedQuantityChange = $requestedQuantity - $currentReservedQuantity;

            // Update the product quantity table
            $update = $this->sql->update('product_quantity');
            $update->set(array(
                'reserved_quantity' => ($currentQuantity['reserved_quantity'] + $reservedQuantityChange)
            ));
            $update->where->equalTo('product_number', $productNumber);

            $statement = $this->sql->prepareStatementForSqlObject($update);
            $statement->execute();

            // Now we update the product reservation table
            // If the requested quantity is 0, we can just delete the record
            if ($requestedQuantity == 0) {
                $action = $this->sql->delete('product_reservation');
                $action->where->equalTo('product_number', $productNumber);
                $action->where->equalTo('product_reservation_token_id', $reservationTokenId);
            } else {
                // Check if we already have a reservation, if not we insert instead
                if ($currentReservedQuantity == 0) {
                    $action = $this->sql->insert('product_reservation');
                    $action->values(array(
                        'product_number' => $productNumber,
                        'reserved_quantity' => $requestedQuantity,
                        'product_reservation_token_id' => $reservationTokenId
                    ));
                } else {
                    $action = $this->sql->update('product_reservation');
                    $action->set(array(
                        'reserved_quantity' => ($currentReservedQuantity + $reservedQuantityChange),
                    ));
                    $action->where->equalTo('product_number', $productNumber);
                    $action->where->equalTo('product_reservation_token_id', $reservationTokenId);
                }
            }

            $statement = $this->sql->prepareStatementForSqlObject($action);
            $statement->execute();

            $connection->commit();

            return true;

        } catch (\Exception $e) {
            $connection->rollback();
            return false;
        }
    }

    /**
     * Increases the reserved quantity of a product for a reservation token.
     *
     * @param string $productNumber the product number to set the reserved quantity of
     * @param int $requestedQuantity the quantity of the product that has been requested to increase the reservation by
     * @param string $reservationToken the reservation token to make the reservation under
     * @param QuantityMapper $quantityMapper the QuantityMapper class, used to get current stock levels
     *
     * @return boolean true if the reservation has been made successfully, or false if there isn't sufficient stock to fulfil the request
     */
    public function increaseReservationQuantity($productNumber, $requestedQuantity, $reservationToken, QuantityMapper $quantityMapper) {
        // Start transaction
        $connection = $this->sql->getAdapter()->getDriver()->getConnection();
        $connection->beginTransaction();

        try {
            // Get current reserved quantity
            $currentReservedQuantity = $this->getReservedQuantityByProductNumberAndReservationToken($productNumber, $reservationToken);

            // Get current quantity
            $currentQuantity = $quantityMapper->getQuantityByProductNumber($productNumber);

            // Get reservation token ID
            $reservationTokenId = $this->getReservationTokenIdByReservationToken($reservationToken);

            // Update the product quantity table
            $update = $this->sql->update('product_quantity');
            $update->set(array(
                'reserved_quantity' => ($currentQuantity['reserved_quantity'] + $requestedQuantity)
            ));
            $update->where->equalTo('product_number', $productNumber);

            $statement = $this->sql->prepareStatementForSqlObject($update);
            $statement->execute();

            // Now we update the product reservation table
            // Check if we already have a reservation, if not we insert instead
            if ($currentReservedQuantity == 0) {
                $action = $this->sql->insert('product_reservation');
                $action->values(array(
                    'product_number' => $productNumber,
                    'reserved_quantity' => $requestedQuantity,
                    'product_reservation_token_id' => $reservationTokenId
                ));
            } else {
                $action = $this->sql->update('product_reservation');
                $action->set(array(
                    'reserved_quantity' => ($currentReservedQuantity + $requestedQuantity),
                ));
                $action->where->equalTo('product_number', $productNumber);
                $action->where->equalTo('product_reservation_token_id', $reservationTokenId);
            }

            $statement = $this->sql->prepareStatementForSqlObject($action);
            $statement->execute();

            $connection->commit();

            return true;

        } catch (\Exception $e) {
            $connection->rollback();
            return false;
        }
    }

    /**
     * Checks out a reservation, reducing stock levels by the appropriate amounts.
     *
     * @param string $reservationToken the reservation token that the reservations were made under
     * @param QuantityMapper $quantityMapper the QuantityMapper class, used to get current stock levels
     *
     * @return boolean true if the check out was made successfully, or false otherwise
     */
    public function checkOutByToken($reservationToken, QuantityMapper $quantityMapper) {
        // Start transaction
        $connection = $this->sql->getAdapter()->getDriver()->getConnection();
        $connection->beginTransaction();

        try {
            // Get reservation token ID
            $reservationTokenId = $this->getReservationTokenIdByReservationToken($reservationToken);

            // Get all reservations for the token
            $reservations = $this->getAllReservationsByToken($reservationToken);

            foreach ($reservations as $reservation) {
                $productNumber = $reservation['product_number'];
                $reservedQuantity = $reservation['reserved_quantity'];

                // Get current reservation quantity for product
                $currentQuantity = $quantityMapper->getQuantityByProductNumber($productNumber);

                // Update the product quantity table to reduce the reserved quantity
                $update = $this->sql->update('product_quantity');
                $update->set(array(
                    'total_quantity' => ($currentQuantity['total_quantity'] - $reservedQuantity),
                    'reserved_quantity' => ($currentQuantity['reserved_quantity'] - $reservedQuantity),
                ));
                $update->where->equalTo('product_number', $productNumber);

                $statement = $this->sql->prepareStatementForSqlObject($update);
                $statement->execute();

                // Remove all product reservations for token
                $delete = $this->sql->delete('product_reservation');
                $delete->where->equalTo('product_reservation_token_id', $reservationTokenId);

                $statement = $this->sql->prepareStatementForSqlObject($delete);
                $statement->execute();
            }

            // Set token as expiration processed
            $this->setReservationTokenExpirationProcessed($reservationTokenId);

            $connection->commit();

            return true;

        } catch (\Exception $e) {
            $connection->rollback();
            return false;
        }
    }

    /**
     * Removes all reservations that have been made under a given reservation token and makes the reserved stock available once again.
     *
     * @param string $reservationToken the reservation token that the reservations were made under
     * @param QuantityMapper $quantityMapper the QuantityMapper class, used to get current stock levels
     *
     * @return boolean true if the reservations were cleared successfully, or false otherwise
     */
    public function clearReservationsByToken($reservationToken, QuantityMapper $quantityMapper) {
        // Start transaction
        $connection = $this->sql->getAdapter()->getDriver()->getConnection();
        $connection->beginTransaction();

        try {
            // Get reservation token ID
            $reservationTokenId = $this->getReservationTokenIdByReservationToken($reservationToken);

            // Get all reservations for the token
            $reservations = $this->getAllReservationsByToken($reservationToken);

            foreach ($reservations as $reservation) {
                $productNumber = $reservation['product_number'];
                $reservedQuantity = $reservation['reserved_quantity'];

                // Get current reservation quantity for product
                $currentQuantity = $quantityMapper->getQuantityByProductNumber($productNumber);

                // Update the product quantity table to reduce the reserved quantity
                $update = $this->sql->update('product_quantity');
                $update->set(array(
                    'reserved_quantity' => ($currentQuantity['reserved_quantity'] - $reservedQuantity)
                ));
                $update->where->equalTo('product_number', $productNumber);

                $statement = $this->sql->prepareStatementForSqlObject($update);
                $statement->execute();

                // Remove all product reservations for token
                $delete = $this->sql->delete('product_reservation');
                $delete->where->equalTo('product_reservation_token_id', $reservationTokenId);

                $statement = $this->sql->prepareStatementForSqlObject($delete);
                $statement->execute();
            }

            // Set token as expiration processed
            $this->setReservationTokenExpirationProcessed($reservationTokenId);

            $connection->commit();

            return true;
        } catch (\Exception $e) {
            $connection->rollback();
            return false;
        }
    }

    /**
     * Gets a list of products and their reserved quantities for a given reservation token.
     *
     * @param string $reservationToken the reservation token that the reservations were made under
     *
     * @return array possibly empty list of all product numbers (product_number) and reserved quantities (reserved_quantity)
     */
    public function getAllReservationsByToken($reservationToken) {
        $select = $this->sql->select('product_reservation');
        $select->columns(array('product_number', 'reserved_quantity'));
        $select->join('product_reservation_token', 'product_reservation_token.id = product_reservation.product_reservation_token_id', array());
        $select->where->equalTo('product_reservation_token.token', $reservationToken);
        $select->forUpdate = true;

        $statement = $this->sql->prepareStatementForSqlObject($select);
        $results = $statement->execute();

        $reservations = array();

        foreach ($results as $r) {
            $reservations[] = $r;
        }

        return $reservations;
    }

    /**
     * Gets the database ID for a reservation token
     *
     * @param string $reservationToken the reservation token to get the database ID for
     *
     * @throws \RuntimeException thrown if the given reservation token doesn't exist
     *
     * @return int the database ID of the reservation token
     */
    private function getReservationTokenIdByReservationToken($reservationToken) {
        $select = $this->sql->select('product_reservation_token');
        $select->where->equalTo('token', $reservationToken);

        $statement = $this->sql->prepareStatementForSqlObject($select);
        $results = $statement->execute()->current();

        if (!$results) {
            throw new \RuntimeException('Reservation Token not found');
        }

        return $results['id'];
    }

    /**
     * Sets flag in database that an expired reservation token has been processed and all stock that was reserved for the given reservation has been released back for re-use.
     *
     * @param int $reservationTokenId the database ID of the reservation token to me parked as processed
     */
    private function setReservationTokenExpirationProcessed($reservationTokenId) {
        $update = $this->sql->update('product_reservation_token');
        $update->set(array(
            'expiration_processed' => 1
        ));
        $update->where->equalTo('id', $reservationTokenId);

        $statement = $this->sql->prepareStatementForSqlObject($update);
        $statement->execute();
    }

    /**
     * Generates a new, unique token to be used for reservations.
     *
     * @return array details of the token (token) and the date and time that it expires (token_expiry) in the format YYYY-MM-DD HH:MM:SS
     */
    public function generateToken() {
        do {
            // Generate a random token
            $token = $this->generateRandomToken();

            // Check if the token already exists in the database
            $select = $this->sql->select('product_reservation_token');
            $select->where->equalTo('token', $token);

            $statement = $this->sql->prepareStatementForSqlObject($select);
            $results = $statement->execute();

        } while ( $results->count() > 0 );

        // Calculate token expiry
        $tokenExpiry = date('Y-m-d H:i:s', strtotime('+' . $this->TOKEN_EXPIRY_IN_MINUTES . 'minutes'));

        $insert = $this->sql->insert('product_reservation_token');
        $insert->values(array(
            'token' => $token,
            'token_expiry' => $tokenExpiry,
            'expiration_processed' => 0
        ));

        $statement = $this->sql->prepareStatementForSqlObject($insert);
        $statement->execute();

        return array(
            'token' => $token,
            'token_expiry' => $tokenExpiry
        );
    }

    /**
     * Internal helper function for generating a random string.
     *
     * @return string the randomly generated string
     */
    private function generateRandomToken() {
        return bin2hex(openssl_random_pseudo_bytes(64));
    }
}

?>